from copy import deepcopy
import json

import cv2
import numpy as np
from pydantic import ValidationError
import pytest

from rvimage.collection_types import BboxAnnos, BrushAnnos, Canvas
from rvimage.converters import (
    decode_bytes_into_rgbarray,
    extract_polys_from_mask,
    fill_polys_on_mask,
    mask_to_rle,
    rle_to_mask,
)
from rvimage.domain import BbF, BbI, Point, Poly


def test_rle():
    im_mask = np.zeros((10, 10), dtype=np.uint8)
    im_mask[0:2, 0:5] = 1

    rle = mask_to_rle(im_mask)
    assert rle == [0, 5, 5, 5, 85]
    im_mask_converted = rle_to_mask(rle, 1, im_mask)
    assert np.array_equal(im_mask, im_mask_converted)


def test_polygon():
    im_mask = np.zeros((10, 10), dtype=np.uint8)
    polygons = [
        [
            Point(x=0, y=0),
            Point(x=5, y=0),
            Point(x=5, y=5),
            Point(x=2, y=7),
            Point(x=0, y=5),
        ]
    ]
    value = 1

    mask = fill_polys_on_mask(polygons, value, im_mask, abs_coords_input=True)
    assert np.sum(mask) > 0

    polygons_converted = extract_polys_from_mask(mask, abs_coords_output=True)
    assert len(polygons_converted) > 0
    im_mask = np.zeros((10, 10), dtype=np.uint8)
    mask_converted = fill_polys_on_mask(polygons, value, im_mask, abs_coords_input=True)

    assert np.allclose(mask, mask_converted), "Masks are not equal after conversion"


def test_validation():
    annos = {
        "elts": [],
        "cat_idxs": [],
        "selected_mask": [],
    }
    BboxAnnos.model_validate(annos)
    annos = {
        "elts": [{"BB": {"x": 0.0, "y": 0.0, "w": 5.0, "h": 5.0}}],
        "cat_idxs": [1],
        "selected_mask": [False],
    }
    BboxAnnos.model_validate(annos)


def test_inbox():
    bb1 = BbI(x=271, y=192, w=1014, h=86)
    bb2 = BbI(x=0, y=190, w=1500, h=100)
    assert bb1 in bb2

    with open("../rvimage/resources/test_data/rvprj_v4-0.json", "r") as f:
        data_loaded = json.load(f)

    def get_data(tool):
        for d, _ in data_loaded["tools_data_map"][tool]["specifics"][tool][
            "annotations_map"
        ].values():
            yield d

    data = get_data("Bbox")
    for i, bbox_data in enumerate(data):
        annos = BboxAnnos.model_validate(bbox_data)
        if i == 0:
            annos_ = deepcopy(annos)
            assert len(annos_.elts) == 4
            assert len(annos_.cat_idxs) == 4
            assert len(annos_.selected_mask) == 4
            assert len(list(annos.bbs())) == 4
            assert len(list(annos.bbs(cat_idx=[0]))) == 4
            assert len(list(annos.bbs(cat_idx=[1]))) == 0

            annos_.keep_only_inbox_annos(
                [
                    BbF(x=550.70, y=1300.28, w=1455, h=339),
                    BbF(x=105.72, y=416.41, w=327.09, h=932.01),
                ]
            )
            assert len(annos_.elts) == 2
            assert len(annos_.cat_idxs) == 2
            assert len(annos_.selected_mask) == 2
            assert len(list(annos_.bbs())) == 2

            annos.remove_inbox_annos(
                [
                    BbF(x=550.70, y=1300.28, w=1455, h=339),
                    BbF(x=105.72, y=416.41, w=327.09, h=932.01),
                ]
            )
            assert len(annos.elts) == 2
            assert len(annos.cat_idxs) == 2
            assert len(annos.selected_mask) == 2
            assert len(list(annos.bbs())) == 2

    data = get_data("Brush")
    for i, brush_data in enumerate(data):
        annos = BrushAnnos.model_validate(brush_data)
        if i == 0:
            annos_ = deepcopy(annos)
            assert len(annos_.elts) == 4
            assert len(annos_.cat_idxs) == 4
            assert len(annos_.selected_mask) == 4
            assert len(list(annos_.bbs())) == 4
            assert len(list(annos_.bbs(cat_idx=[0, 6]))) == 4
            assert len(list(annos_.bbs(cat_idx=[1, 3]))) == 0

            annos_.keep_only_inbox_annos([bb2])
            assert len(annos_.elts) == 1
            assert len(annos_.cat_idxs) == 1
            assert len(annos_.selected_mask) == 1
            assert len(list(annos_.bbs())) == 1

            annos.remove_inbox_annos([bb2])
            assert len(annos.elts) == 3
            assert len(annos.cat_idxs) == 3
            assert len(annos.selected_mask) == 3
            assert len(list(annos.bbs())) == 3


def test_from_mask():
    resulting_mask = np.zeros((64, 32), dtype=np.uint8)
    resulting_mask[31:40, 21:30] = 1
    BboxAnnos.from_mask(resulting_mask, 0)
    annos = BrushAnnos.from_mask(resulting_mask, 0)
    reconstructed_mask = np.zeros_like(resulting_mask)
    annos.fill_mask(reconstructed_mask, 0)
    assert np.array_equal(resulting_mask, reconstructed_mask), (
        "Reconstructed mask does not match the original mask"
    )


def test_len_validator():
    bb1 = BbF(x=11, y=0, w=10, h=20)
    with pytest.raises(ValidationError):
        BboxAnnos(elts=[bb1], cat_idxs=[], selected_mask=[])
    BboxAnnos(elts=[bb1], cat_idxs=[0], selected_mask=[True])
    rle = mask_to_rle(np.zeros((10, 20)))
    canvas = Canvas(rle=rle, bb=bb1.to_bbi(), intensity=1)
    with pytest.raises(ValidationError):
        BrushAnnos(elts=[canvas], cat_idxs=[], selected_mask=[])
    with pytest.raises(ValidationError):
        BrushAnnos(elts=[], cat_idxs=[1], selected_mask=[])
    BrushAnnos(elts=[], cat_idxs=[], selected_mask=[])
    BrushAnnos(elts=[canvas], cat_idxs=[1], selected_mask=[True])


def test_decode_image():
    bytes = open("../rvimage/resources/rvimage-logo.png", "rb").read()
    im_decoded = decode_bytes_into_rgbarray(bytes)
    im_read = cv2.imread("../rvimage/resources/rvimage-logo.png", cv2.IMREAD_COLOR)
    im_read = cv2.cvtColor(im_read, cv2.COLOR_BGR2RGB)
    assert im_decoded.shape == im_read.shape, (
        "Decoded image shape does not match read image shape"
    )
    assert np.array_equal(im_decoded, im_read), (
        "Decoded image does not match read image"
    )


def test_bb_rowcolinterface():
    def test(bb: BbI | BbF):
        assert bb.c_min == bb.x
        assert bb.r_min == bb.y
        assert bb.r_max == bb.y + bb.h
        assert bb.c_max == bb.x + bb.w
        assert bb.height == bb.h
        assert bb.width == bb.w
        if isinstance(bb, BbI):
            y_slice, x_slice = bb.slices
            assert y_slice == slice(bb.r_min, bb.r_max)
            assert x_slice == slice(bb.c_min, bb.c_max)

    test(BbF(x=0, y=0, w=10, h=20))
    test(BbI(x=0, y=0, w=1, h=2))
    test(BbF(x=0.5, y=0.76, w=10.8, h=20.22))
    test(BbI(x=1, y=120, w=11, h=21))


def test_intersect():
    bb1 = BbF(x=0, y=0, w=10, h=20)
    bb2 = BbI(x=0, y=0, w=10, h=20)
    # match the float box, not the int box
    assert bb1.intersect(bb2) == bb1
    assert bb1.intersect(bb2) != bb2

    # no overlap
    bb1 = BbF(x=11, y=0, w=10, h=20)
    bb4 = BbI(x=0, y=0, w=10, h=20)
    assert bb1.intersect(bb2) is None

    bb1 = BbF(x=5, y=6, w=10, h=20)
    bb3 = BbF(x=0, y=0, w=10, h=20)
    assert bb1.intersect(bb3) == BbF(x=5, y=6, w=5, h=14)

    bb1 = BbI(x=15, y=6, w=100, h=20)
    bb2 = BbI(x=20, y=10, w=100, h=120)
    assert bb1.intersect(bb2) == BbI(x=20, y=10, w=95, h=16)
    assert bb1.find_max_overlap_bb([bb1, bb2]) == bb1
    bb1 = BbI(x=15, y=6, w=100, h=20)
    bb2 = BbI(x=20, y=10, w=100, h=120)
    bbf = BbF(x=20, y=10, w=100, h=120)
    assert bbf.find_max_overlap_bb([bb1, bb2]) == bb2
    assert bb1.find_max_overlap_bb([bb2, bb3, bb4]) == bb2
    assert bb1.find_max_overlap_bb([bb4]) is None


def test_bb_conversion():
    bbi = BbI(x=0, y=0, w=10, h=20)
    assert BbF.from_bbi(bbi).to_bbi() == bbi

    bbf = BbF(x=0.04, y=0.4, w=9.54, h=20.3)
    assert bbf.to_bbi() == bbi
    assert bbf.to_bbi() == bbi.to_bbi()


def test_singleelt():
    bbf1 = BbF(x=0.04, y=0.4, w=9.54, h=20.3)
    bbf2 = BbF(x=10.47, y=0.7, w=3.54, h=7.3)
    bbf3 = BbF(x=8.14, y=6.2, w=8.54, h=10.3)
    bbf4 = BbF(x=8.14, y=6.2, w=8.54, h=10.3)
    annos = BboxAnnos.from_elt(bbf1, 0)
    annos.append_elt(bbf2, cat_idx=1)
    annos.append_elt(bbf3, cat_idx=3)
    annos.append_elt(bbf4, cat_idx=3)
    annos_ref = BboxAnnos(
        elts=[bbf1, bbf2, bbf3], cat_idxs=[0, 1, 3], selected_mask=[False, False, False]
    )
    assert annos == annos_ref

    # with a different category, it is not a duplicate
    annos.append_elt(bbf4, cat_idx=4)
    annos_ref = BboxAnnos(
        elts=[bbf1, bbf2, bbf3, bbf4],
        cat_idxs=[0, 1, 3, 4],
        selected_mask=[False, False, False, False],
    )
    assert annos == annos_ref
    poly1 = Poly.from_points(
        [Point(x=1.2, y=1.3), Point(x=2.2, y=1.6), Point(x=11.2, y=10.3)]
    )
    annos.append_elt(poly1, cat_idx=4)
    annos_ref = BboxAnnos(
        elts=[bbf1, bbf2, bbf3, bbf4, poly1],
        cat_idxs=[0, 1, 3, 4, 4],
        selected_mask=[False, False, False, False, False],
    )
    assert annos == annos_ref

    annos.extend(annos_ref)
    annos_ref = BboxAnnos(
        elts=[bbf1, bbf2, bbf3, bbf4, poly1],
        cat_idxs=[0, 1, 3, 4, 4],
        selected_mask=[
            False,
            False,
            False,
            False,
            False,
        ],
    )
    assert annos == annos_ref
    annos_ = BboxAnnos.from_elt(bbf1, 6)
    annos_.append_elt(bbf2, 7, True)
    annos.extend(annos_)
    annos_ref = BboxAnnos(
        elts=[bbf1, bbf2, bbf3, bbf4, poly1, bbf1, bbf2],
        cat_idxs=[0, 1, 3, 4, 4, 6, 7],
        selected_mask=[False, False, False, False, False, False, True],
    )
    assert annos == annos_ref


def test_equals():
    bbf1 = BbF(x=0.04, y=0.4, w=9.54, h=20.3)
    bbf2 = BbF(x=10.47, y=0.7, w=3.54, h=7.3)
    bbf3 = BbF(x=8.14, y=6.2, w=8.54, h=10.3)
    bbf4 = BbF(x=8.14, y=6.2, w=8.54, h=10.3)
    bbf5 = BbF(x=8.14 + 1e-9, y=6.2, w=8.54, h=10.3)
    assert not bbf1.equals(bbf2)
    assert bbf3.equals(bbf4)
    assert bbf3.equals(bbf5)

    poly1 = Poly.from_points(
        [Point(x=1.2, y=1.3), Point(x=2.2, y=1.6), Point(x=11.2, y=10.3)]
    )
    poly2 = Poly.from_points(
        [Point(x=1.2, y=1.3), Point(x=2.2, y=1.6), Point(x=11.2, y=10.3)]
    )
    poly3 = Poly.from_points(
        [Point(x=2.2, y=1.3), Point(x=4.2, y=4.6), Point(x=11.2, y=19.3)]
    )
    poly4 = Poly.from_points(
        [Point(x=2.2, y=1.3 + 1e-9), Point(x=4.2, y=4.6), Point(x=11.2, y=19.3)]
    )
    assert poly1.equals(poly2)
    assert not poly1.equals(poly3)
    assert poly3.equals(poly4)


if __name__ == "__main__":
    test_len_validator()
    test_equals()
    test_singleelt()
    test_bb_conversion()
    test_intersect()
    test_inbox()
    test_bb_rowcolinterface()
    test_from_mask()
    test_decode_image()
    test_validation()
    test_rle()
    test_polygon()



from __future__ import annotations

from collections.abc import Callable, Sequence
from typing import Protocol, Self, TypeVar
from collections.abc import Iterable

import numpy as np
from pydantic import BaseModel, model_serializer, model_validator

from rvimage.converters import (
    extract_polys_from_mask,
    fill_bbs_on_mask,
    fill_polys_on_mask,
    mask_to_rle,
    rle_to_mask,
)
from rvimage.domain import BbF, BbI, Poly, find_ccs

B = TypeVar("B", BbI, BbF)
T = TypeVar("T")


def _inbox_inds(
    container_boxes: Sequence[BbF | BbI],
    keep_or_remove: Sequence[T],
    convert_to_box: Callable[[T], B],
):
    return [
        i
        for i, elt in enumerate(keep_or_remove)
        if any(convert_to_box(elt) in cb for cb in container_boxes)
    ]


def _outofbox_inds(
    container_boxes: Sequence[BbF | BbI],
    keep_or_remove: Sequence[T],
    convert_to_box: Callable[[T], B],
):
    return [
        i
        for i, elt in enumerate(keep_or_remove)
        if all(convert_to_box(elt) not in cb for cb in container_boxes)
    ]


class _AnnotationProtocol(Protocol):
    elts: list
    cat_idxs: list[int]
    selected_mask: list[bool]


def _keep_inds(obj: _AnnotationProtocol, inds: list[int]):
    obj.elts = [obj.elts[i] for i in inds]
    obj.cat_idxs = [obj.cat_idxs[i] for i in inds]
    obj.selected_mask = [obj.selected_mask[i] for i in inds]


class Labelinfo(BaseModel):
    new_label: str
    labels: list[str]
    colors: list[list[int]]
    cat_ids: list[int]
    cat_idx_current: int
    show_only_current: bool


def _bbox_elt_to_bb(elt: BbF | Poly) -> BbF:
    return elt if isinstance(elt, BbF) else elt.enclosing_bb


class BboxAnnos(BaseModel):
    elts: list[BbF | Poly]
    cat_idxs: list[int]
    selected_mask: list[bool]

    @model_validator(mode="before")
    @classmethod
    def resolve_bb_poly(cls, data: dict) -> dict:
        # remove the type-info from the dict
        if len(data["elts"]) > 0 and isinstance(data["elts"][0], dict):
            data["elts"] = [next(v for v in d.values()) for d in data["elts"]]
        return data

    @model_validator(mode="after")
    def check_len(self) -> "Self":
        assert len(self.elts) == len(self.cat_idxs) == len(self.selected_mask)
        return self

    @model_serializer()
    def serialize_model(self):
        elts = [
            {"BB": elt.model_dump()}
            if isinstance(elt, BbF)
            else {"Poly": elt.model_dump()}
            for elt in self.elts
        ]
        return {
            "cat_idxs": self.cat_idxs,
            "selected_mask": self.selected_mask,
            "elts": elts,
        }

    @classmethod
    def from_mask(cls, mask: np.ndarray, cat_idx: int) -> "BboxAnnos":
        """
        Create BboxAnnos from a binary mask.
        """
        polys = extract_polys_from_mask(mask, abs_coords_output=True)
        cat_idxs = [cat_idx] * len(polys)

        return cls(
            elts=[Poly.from_points(points) for points in polys],
            cat_idxs=cat_idxs,
            selected_mask=[False] * len(polys),
        )

    def extend(self, other: Self | None):
        """
        Extend the current BboxAnnos with another BboxAnnos.
        """
        if other is None:
            return self
        for o_elt, o_cat_idx, o_selected in zip(
            other.elts, other.cat_idxs, other.selected_mask
        ):
            self.append_elt(o_elt, o_cat_idx, o_selected)

    def append_elt(
        self, elt: BbI | BbF | Poly, cat_idx: int, selected_mask: bool = False
    ):
        if isinstance(elt, BbI):
            elt = BbF.from_bbi(elt)
        is_duplicate = any(
            elt.equals(elt_i) and cat_idx == cat_idx_i
            for elt_i, cat_idx_i in zip(self.elts, self.cat_idxs)
        )
        if not is_duplicate:
            self.elts.append(elt)
            self.cat_idxs.append(cat_idx)
            self.selected_mask.append(selected_mask)

    @classmethod
    def from_elt(cls, elt: BbI | BbF | Poly, cat_idx: int) -> "BboxAnnos":
        if isinstance(elt, BbI):
            elt = BbF.from_bbi(elt)
        return cls(elts=[elt], cat_idxs=[cat_idx], selected_mask=[False])

    def fill_mask(self, im_mask: np.ndarray, cat_idx: int, value: int = 1):
        fill_polys_on_mask(
            polygons=(
                elt.points
                for elt, cat_idx_ in zip(self.elts, self.cat_idxs)
                if cat_idx == cat_idx_ and isinstance(elt, Poly)
            ),
            value=value,
            im_mask=im_mask,
            abs_coords_input=True,
        )
        fill_bbs_on_mask(
            bbs=(
                elt
                for elt, cat_idx_ in zip(self.elts, self.cat_idxs)
                if cat_idx == cat_idx_ and isinstance(elt, BbF)
            ),
            value=value,
            im_mask=im_mask,
            abs_coords_input=True,
        )

    def bbs(self, cat_idx: list[int] | None = None) -> Iterable[BbF]:
        return (
            elt if isinstance(elt, BbF) else elt.enclosing_bb
            for elt, cat_idx_ in zip(self.elts, self.cat_idxs)
            if cat_idx is None or cat_idx_ in cat_idx
        )

    def keep_only_inbox_annos(self, bbs: Sequence[BbF | BbI]):
        """Keep all annotations whose bounding box is contained in one of the passed bbs"""
        inds = _inbox_inds(
            bbs,
            self.elts,
            _bbox_elt_to_bb,
        )
        _keep_inds(self, inds)

    def remove_inbox_annos(self, bbs: Sequence[BbF | BbI]):
        """Remove all annotations whose bounding box is contained in one of the passed bbs"""
        inds = _outofbox_inds(
            bbs,
            self.elts,
            lambda elt: elt if isinstance(elt, BbF) else elt.enclosing_bb,
        )
        _keep_inds(self, inds)

    def find_max_overlap_bb_with_zoombox(self, zb: BbI, cat_idxs: list[int]) -> BbI:
        if len(self.elts) > 0:
            max_ol_bb = zb.find_max_overlap_bb(self.bbs(cat_idxs))
            if max_ol_bb is None:
                return zb
            else:
                return max_ol_bb.to_bbi()
        else:
            return zb


class BboxData(BaseModel):
    annos: BboxAnnos
    labelinfo: Labelinfo


class Canvas(BaseModel):
    rle: list[int]
    bb: BbI
    intensity: float


def _brush_elt_to_bb(elt: Canvas) -> BbI:
    return elt.bb


class BrushAnnos(BaseModel):
    elts: list[Canvas]
    cat_idxs: list[int]
    selected_mask: list[bool]

    @model_validator(mode="after")
    def check_len(self) -> "Self":
        assert len(self.elts) == len(self.cat_idxs) == len(self.selected_mask)
        return self

    @classmethod
    def from_mask(cls, im_mask: np.ndarray, cat_idx: int) -> "BrushAnnos":
        """
        Create BrushAnnos from a binary mask.
        """
        ccs, _ = find_ccs(im_mask)
        rles = [mask_to_rle(cc.im) for cc in ccs]

        return BrushAnnos(
            elts=[
                Canvas(rle=rle, bb=cc.bb, intensity=1.0) for rle, cc in zip(rles, ccs)
            ],
            cat_idxs=[cat_idx] * len(ccs),
            selected_mask=[False] * len(ccs),
        )

    def fill_mask(self, im_mask: np.ndarray, cat_idx: int, value: int = 1):
        """Create a binary mask from all brush annotations of the given category

        Args:
            im_mask: output mask to write on
            cat_idx: index of category to be written
            value: image intensity of masked pixels
        """
        for elt, cat_idx_ in zip(self.elts, self.cat_idxs):
            if cat_idx == cat_idx_:
                im_bb_mask = rle_to_mask(elt.rle, value=value, mask=elt.bb)
                im_mask[elt.bb.slices][im_bb_mask > 0] = im_bb_mask[im_bb_mask > 0]

    def extend(self, other: Self | None):
        """
        Extend the current BrushAnnos with another BrushAnnos.
        """
        if other is None:
            return self
        self.elts += other.elts
        self.cat_idxs += other.cat_idxs
        self.selected_mask += other.selected_mask

    def bbs(self, cat_idx: list[int] | None = None) -> Iterable[BbI]:
        return (
            _brush_elt_to_bb(elt)
            for elt, cat_idx_ in zip(self.elts, self.cat_idxs)
            if cat_idx is None or cat_idx_ in cat_idx
        )

    def keep_only_inbox_annos(self, bbs: Sequence[BbI | BbF]):
        """Keep all annotations whose bounding box is contained in of the passed bbs"""
        inds = _inbox_inds(bbs, self.elts, _brush_elt_to_bb)
        _keep_inds(self, inds)

    def remove_inbox_annos(self, bbs: Sequence[BbI | BbF]):
        """Keep all annotations whose bounding box is contained in of the passed bbs"""
        inds = _outofbox_inds(bbs, self.elts, _brush_elt_to_bb)
        _keep_inds(self, inds)


class BrushData(BaseModel):
    annos: BrushAnnos
    labelinfo: Labelinfo


class InputAnnotationData(BaseModel):
    bbox: BboxData | None
    brush: BrushData | None


class OutputAnnotationData(BaseModel):
    bbox: BboxAnnos | None
    brush: BrushAnnos | None



from __future__ import annotations

import math
from typing import Generic, Iterable, TypeVar

import numpy as np
from pydantic import BaseModel
from scipy.ndimage import find_objects
from scipy.ndimage import label as scpiy_label


T = TypeVar("T", int, float)


class _RowColMixin(Generic[T]):
    @property
    def r_min(self) -> T:
        return self.y  # type: ignore[attr-defined]

    @property
    def r_max(self) -> T:
        return self.y + self.h  # type: ignore[attr-defined]

    @property
    def c_min(self) -> T:
        return self.x  # type: ignore[attr-defined]

    @property
    def c_max(self) -> T:
        return self.x + self.w  # type: ignore[attr-defined]

    @property
    def width(self) -> T:
        return self.w  # type: ignore[attr-defined]

    @property
    def height(self) -> T:
        return self.h  # type: ignore[attr-defined]

    @property
    def area(self) -> T:
        return self.height * self.width

    @classmethod
    def from_rowcols(cls, r_min: T, c_min: T, r_max: T, c_max: T):
        x = c_min
        y = r_min
        w = c_max - c_min
        h = r_max - r_min
        return cls(x=x, y=y, h=h, w=w)  # type: ignore[attr-defined]


class _ContainsMixin:
    def __contains__(self, other: BbI | BbF) -> bool:
        return (
            other.c_min >= self.c_min  # type: ignore[attr-defined]
            and other.c_max <= self.c_max  # type: ignore[attr-defined]
            and other.r_min >= self.r_min  # type: ignore[attr-defined]
            and other.r_max <= self.r_max  # type: ignore[attr-defined]
        )

    def intersect(self, other: BbI | BbF) -> BbI | BbF | None:
        c_min = max(self.c_min, other.c_min)  # type: ignore[attr-defined]
        c_max = min(self.c_max, other.c_max)  # type: ignore[attr-defined]
        r_min = max(self.r_min, other.r_min)  # type: ignore[attr-defined]
        r_max = min(self.r_max, other.r_max)  # type: ignore[attr-defined]
        if c_min < c_max and r_min < r_max:
            if (
                isinstance(c_min, int)
                and isinstance(c_max, int)
                and isinstance(r_min, int)
                and isinstance(r_max, int)
            ):
                return BbI.from_rowcols(r_min, c_min, r_max, c_max)
            else:
                return BbF.from_rowcols(r_min, c_min, r_max, c_max)
        else:
            # boxes don't overlap
            return None

    def find_max_overlap_bb(self, others: Iterable[BbI | BbF]) -> BbI | BbF | None:
        """Finds the last bounding box from others with maximum overlap with self"""
        max_overlap = 0
        max_overlap_bb = None
        for other_bb in others:
            inter = self.intersect(other_bb)
            ol = inter and inter.area
            if ol is not None and ol >= max_overlap:
                max_overlap = ol
                max_overlap_bb = other_bb
        return max_overlap_bb


class BbI(BaseModel, _RowColMixin[int], _ContainsMixin):
    x: int
    y: int
    w: int
    h: int

    @classmethod
    def from_slices(cls, slices: tuple[slice, slice]):
        x = slices[1].start
        y = slices[0].start
        w = slices[1].stop - x
        h = slices[0].stop - y
        return cls(x=x, y=y, w=w, h=h)

    @property
    def slices(self) -> tuple[slice, slice]:
        """
        Returns the slices for indexing a numpy array.
        """
        return slice(self.y, self.y + self.h), slice(self.x, self.x + self.w)

    def to_bbi(self):
        return self


class BbF(BaseModel, _RowColMixin[float], _ContainsMixin):
    x: float
    y: float
    w: float
    h: float

    def to_bbi(self) -> BbI:
        """
        Convert to integer bounding box.
        """
        return BbI(
            x=int(np.round(self.x)),
            y=int(np.round(self.y)),
            w=int(np.round(self.w)),
            h=int(np.round(self.h)),
        )

    @classmethod
    def from_bbi(cls, bbi: BbI):
        return cls(x=bbi.x, y=bbi.y, w=bbi.w, h=bbi.h)

    def scale(self, scale_x: float, scale_y) -> "BbF":
        """
        Scale the bounding box by a factor.
        """
        return BbF(
            x=self.x * scale_x,
            y=self.y * scale_y,
            w=self.w * scale_x,
            h=self.h * scale_y,
        )

    def equals(self, other: "BbF | Poly", tolerance: float = 1e-8) -> bool:
        if isinstance(other, Poly):
            return False
        if abs(self.x - other.x) > tolerance:
            return False
        if abs(self.y - other.y) > tolerance:
            return False
        if abs(self.w - other.w) > tolerance:
            return False
        if abs(self.h - other.h) > tolerance:
            return False
        return True


class Point(BaseModel):
    x: float
    y: float


class Poly(BaseModel):
    points: list[Point]
    enclosing_bb: BbF

    @classmethod
    def from_points(cls, points: list[Point]) -> "Poly":
        return cls(points=points, enclosing_bb=enclosing_bb(points))

    def equals(self, other: "Poly | BbF", tolerance: float = 1e-8) -> bool:
        if isinstance(other, BbF):
            return False
        if not self.enclosing_bb.equals(other.enclosing_bb):
            return False
        for p1, p2 in zip(self.points, other.points):
            if abs(p1.x - p2.x) > tolerance or abs(p1.y - p2.y) > tolerance:
                return False
        return True


def enclosing_bb(points: list[Point]) -> BbF:
    points = points
    min_x = math.inf
    min_y = math.inf
    max_x = -math.inf
    max_y = -math.inf
    for point in points:
        if point.x < min_x:
            min_x = point.x
        if point.y < min_y:
            min_y = point.y
        if point.x > max_x:
            max_x = point.x
        if point.y > max_y:
            max_y = point.y
    return BbF(x=min_x, y=min_y, w=max_x + 1 - min_x, h=max_y + 1 - min_y)


class CC:
    """Connected component"""

    def __init__(
        self,
        slices: tuple[slice, slice],
        label: int,
        im: np.ndarray,
        im_labeled: np.ndarray,
    ):
        self.im = im[slices].copy()
        self.im[im_labeled[slices] != label] = 0
        self.slices = slices
        self.bb = BbI.from_slices(slices)
        self.label = label

    def __str__(self):
        return "CC with " + str(self.bb)


def _find_cc_slices(im: np.ndarray):
    im_labeled, n_ccs = scpiy_label(im)  # type: ignore
    return find_objects(im_labeled), im_labeled, n_ccs


def find_ccs(im: np.ndarray) -> tuple[list[CC], np.ndarray]:
    """Find connected components in a binary image.
    Args:
        im: A binary image (2D numpy array) where connected components are to be found.
    Returns:
        A tuple containing:
            - A list of CC objects representing the connected components.
            - A labeled image where each connected component is assigned a unique label.
    """
    cc_slices, im_labeled, _ = _find_cc_slices(im)
    ccs = [CC(slc, i + 1, im, im_labeled) for i, slc in enumerate(cc_slices)]
    return ccs, im_labeled
